@@ -1,8 +1,8 @@
-context("tbl_json: as.tbl_json.character")
+context("tbl_json: as_tbl_json.character")
 
 test_that("correctly parses length(json) == 1", {
   expect_identical(
-    as.tbl_json('{"name": "bob", "age": 32}'),
+    as_tbl_json('{"name": "bob", "age": 32}'),
     tbl_json(
       data.frame(document.id = 1L),
       list(list(name = "bob", age = 32L))
@@ -12,7 +12,7 @@ test_that("correctly parses length(json) == 1", {
 
 test_that("correctly parses length(json) > 1", {
   expect_identical(
-    as.tbl_json(
+    as_tbl_json(
       c('{"name": "bob", "age": 32}',
         '{"name": "susan", "age": 25}')
     ),
@@ -28,7 +28,7 @@ test_that("correctly parses length(json) > 1", {
 
 test_that("correctly parses character(0)", {
   expect_identical(
-    as.tbl_json(character(0)),
+    as_tbl_json(character(0)),
     tbl_json(
       data.frame(document.id = integer(0)),
       list()
@@ -42,7 +42,7 @@ test_that("correctly parses empty objects", {
   names(nl) <- character(0)
 
   expect_identical(
-    as.tbl_json(c('[]', '{}')),
+    as_tbl_json(c('[]', '{}')),
     tbl_json(
       data.frame(document.id = 1L:2L),
       list(list(), nl)
@@ -53,7 +53,7 @@ test_that("correctly parses empty objects", {
 
 test_that("correctly structures an array", {
   expect_identical(
-    as.tbl_json('[{"name": "bob"}, {"name": "susan"}]'),
+    as_tbl_json('[{"name": "bob"}, {"name": "susan"}]'),
     tbl_json(
       data.frame(document.id = 1L),
       list(list(list(name = "bob"), list(name = "susan")))
@@ -61,9 +61,27 @@ test_that("correctly structures an array", {
   )
 })
 
+test_that("[ works with various indexing", {
+  obj <- as_tbl_json(c('{"name": "value"}', '{"name": "other"}')) %>%
+    json_types()
+  # column indexing
+  expect_identical(obj[0], obj[,0])
+  expect_identical(obj[1], obj[,1])
+  expect_identical(obj[-1], obj[,-1])
+  expect_identical(obj[0:2], obj)
+  
+  # no indexing
+  expect_identical(obj, obj[])
+  
+  # row indexing
+  expect_identical(obj[1,], obj[1,1:2])
+  expect_identical(obj[-1,], obj[2,])
+  expect_identical(obj[1:2,], obj)
+})
+
 test_that("throws error on invalid json", {
 
-    expect_error(as.tbl_json(''))
+    expect_error(as_tbl_json(''))
 
   }
 )
@@ -71,7 +89,7 @@ test_that("throws error on invalid json", {
 context("tbl_json: as.character.tbl_json")
 
 inverts_json_test <- function(json) {
-  expect_identical(json, json %>% as.tbl_json %>% as.character)
+  expect_identical(json, json %>% as_tbl_json %>% as.character)
 }
 
 test_that("works for simple cases", {
@@ -104,27 +122,27 @@ test_that("works for worldbank data", {
 })
 
 test_that("throws informative warning message when attr(.,'JSON') is missing", {
-  j <- '{"a": 1, "b": "test"}' %>% as.tbl_json()
-  attr(j,'JSON') <- NULL
+  j <- '{"a": 1, "b": "test"}' %>% as_tbl_json()
+  j[["..JSON"]] <- NULL
   
-  expect_warning(j %>% as.character(),'attr.*JSON.*remove.*tbl_json')
+  expect_warning(j %>% as.character(),'\\.\\.JSON.*remove.*tbl_json')
   expect_identical(suppressWarnings(j %>% as.character()),character())
 })
 
 
-context("as.tbl_json.tbl_json")
+context("as_tbl_json.tbl_json")
 
 test_that('functions as the identity on a simple pipeline', {
    x <- commits %>% gather_array() %>% enter_object('commit') %>% spread_all()
    
    expect_identical(
-     x, as.tbl_json(x)
+     x, as_tbl_json(x)
    )
    
    y <- commits %>% gather_array() %>% gather_object()
    
    expect_identical(
-     y, as.tbl_json(y)
+     y, as_tbl_json(y)
    )
 })
 
@@ -142,21 +160,36 @@ test_that('functions as the identity on a more advanced pipeline', {
   )
   
   expect_identical(
-    x, as.tbl_json(x)
+    x, as_tbl_json(x)
   )
 })
 
+context("as_tbl_json.list")
+
+test_that("list is interpreted as already parsed", {
+  j <- jsonlite::fromJSON('{"a": "b"}')
+  
+  expect_is(as_tbl_json(j), "tbl_json")
+  
+  l <- list(a = "b", c = list("a","b","c"))
+  
+  expect_is(as_tbl_json(l), "tbl_json")
+  expect_identical(as_tbl_json(l)$document.id, c(1L,2L))
+  
+  expect_is(as_tbl_json(list()), "tbl_json")
+})
+
 context("print.tbl_json")
 
 test_that("jsonlite::toJSON works as anticipated", {
-  expect_identical(jsonlite::toJSON(attr(as.tbl_json('"a"'),'JSON')
+  expect_identical(jsonlite::toJSON(json_get(as.tbl_json('"a"'))
                                     , null='null'
                                     , auto_unbox = TRUE) %>% as.character
                    , "[\"a\"]")
 })
 
 test_that("purrr::map_chr works as expected", {
-  a <- attr(as.tbl_json('"a"','JSON'),'JSON') %>% purrr::map_chr(jsonlite::toJSON,
+  a <- json_get(as.tbl_json('"a"','JSON')) %>% purrr::map_chr(jsonlite::toJSON,
                           null = "null",
                           auto_unbox = TRUE)
   
@@ -275,6 +308,16 @@ test_that("works in a pipeline", {
   }
 )
 
+test_that("handles nested lists as a JSON column", {
+  tj <- as_tbl_json('{"name": "value"}')
+  tj_tib <- tibble::as_tibble(json_get_column(tj, "json"))
+  
+  expect_identical(
+    as_tbl_json(tj_tib, json.column = "json"),
+    tj
+    )
+})
+
 test_that("throws an error without json.column specified", {
     expect_error(as.tbl_json(iris))
   }
@@ -295,11 +338,9 @@ test_that("tbl_json constructor works with no data", {
 )
 
 test_that("tbl_json fails if ..JSON is in the names of the data.frame", {
-
-    expect_error(tbl_json(data.frame(..JSON = character(0)), list()))
-
-  }
-)
+  skip("no longer an error")
+  expect_error(tbl_json(data.frame(..JSON = character(0)), list()))
+})
 
 test_that("[ row filtering works with a simple example", {
 
@@ -347,13 +388,41 @@ test_that('handles "drop" like a tbl_df', {
   #expect_warning(is.tbl_json(mydata[,'name',drop=TRUE]),'drop ignored')
 })
 
+test_that("[ keeps column order consistent", {
+  tj <- as_tbl_json('[{"a": "b"}, {"a": "c"}]') %>%
+    gather_array() %>%
+    gather_object()
+  
+  expect_identical(
+    names(tj[1,"document.id"]),
+    c("document.id", "..JSON")
+  )
+  
+  expect_identical(
+    names(tj[2,c("..JSON", "document.id")]),
+    c("..JSON", "document.id")
+  )
+  
+})
+
+test_that("$ leaves tbl_json idempotent", {
+  tj <- as_tbl_json('{"a": "b"}')
+  
+  tj$alt <- 1
+  
+  expect_identical(
+    tj,
+    mutate(tj, alt = alt)
+  )
+})
+
 context('as_tibble') 
 
 test_that('as_tibble drops the JSON attribute and tbl_json class', {
   
   jtidy <- issues %>% gather_array() %>% spread_all()
   
-  expect_identical(attr(dplyr::as_tibble(jtidy),'JSON'),NULL)
+  expect_identical(json_get(dplyr::as_tibble(jtidy)),NULL)
   expect_false('tbl_json' %in% class(dplyr::as_tibble(jtidy)))
 })
 
@@ -366,12 +435,27 @@ test_that('as_data_frame functions like as_tibble', {
     , user.login=jstring('user.login')
   )
   
-  expect_identical(attr(dplyr::as_tibble(jtidy),'JSON'),NULL)
+  expect_identical(json_get(dplyr::as_tibble(jtidy)),NULL)
   expect_false('tbl_json' %in% class(dplyr::as_tibble(jtidy)))
 })
 
 context("tbl_json: dplyr NSE verbs")
 
+test_that("dplyr::group_by works", {
+  skip("wait: vctrs")
+  tj <- as_tbl_json('{"a": "b"}')
+  
+  g1 <- tj %>% group_by(document.id) %>% mutate(a = n())
+  
+  #group_by drops the class today
+  expect_false(inherits(g1, "tbl_json"))
+  
+  expect_error(
+    {tj %>% group_by(..JSON) %>% mutate(b = n())},
+    class = "vctrs_error_subscript_oob"
+  )
+})
+
 test_that("dplyr::filter works with a simple example", {
 
     x <- as.tbl_json(c('{"name": "bob"}', '{"name": "susan"}'))
@@ -460,13 +544,27 @@ test_that("dplyr::mutate works in a more complex pipeline", {
   }
 )
 
+test_that("dplyr::transmute works", {
+  obj <- as.tbl_json(c('{"name": "value"}', '{"name": "string"}'))
+  
+  prep <- obj %>% gather_object %>% append_values_string()
+  
+  use_transmute <- prep %>% transmute(string = paste0(string, "_hi"))
+  
+  expect_is(use_transmute, "tbl_json")
+  expect_identical(nrow(use_transmute), 2L)
+  expect_identical(ncol(use_transmute), 2L)
+  
+  expect_identical(use_transmute$string, c("value_hi", "string_hi"))
+})
+
 test_that("dplyr::slice works", {
 
   new <- '[1, 2, 3]' %>% gather_array %>% dplyr::slice(1:2)
 
   expect_is(new, "tbl_json")
   expect_identical(nrow(new), 2L)
-  expect_identical(length(attr(new, "JSON")), 2L)
+  expect_identical(length(json_get(new)), 2L)
 
 })
 
@@ -476,9 +574,17 @@ test_that('dplyr::select works', {
   f <- json %>% as.tbl_json %>% gather_array %>% spread_all %>%
     dplyr::select(ID=id, object)
   
-  expect_equal(names(f), c('ID','object'))
+  expect_equal(names(f), c('ID','object','..JSON'))
   expect_equal(nrow(f),2)
   expect_is(f,'tbl_json')
+  
+  # Specifically trying to avoid "Adding missing grouping variables: `..JSON`"
+  expect_silent(hm <- as_tbl_json(json) %>% select(document.id))
+  
+  expect_identical(
+    select(f, ..JSON, ID, object),
+    f
+  )
 })
 
 test_that("dplyr::rename works", {
@@ -486,7 +592,7 @@ test_that("dplyr::rename works", {
   new <- '[1, 2, 3]' %>% gather_array %>% dplyr::rename(blah = document.id)
 
   expect_is(new, "tbl_json")
-  expect_identical(names(new), c("blah", "array.index"))
+  expect_identical(names(new), c("blah", "array.index","..JSON"))
 
 })
 
@@ -495,7 +601,7 @@ test_that("dplyr::transmute works", {
   new <- '[1, 2, 3]' %>% gather_array %>% dplyr::transmute(blah = document.id)
 
   expect_is(new, "tbl_json")
-  expect_identical(names(new), "blah")
+  expect_identical(names(new), c("blah", "..JSON"))
 
 })
 
@@ -504,7 +610,7 @@ test_that("dplyr::sample_n works", {
   new <- '[1, 2, 3, 4, 5, 6, 7, 8, 9, 10]' %>% gather_array %>% dplyr::sample_n(2)
 
   expect_is(new, "tbl_json")
-  expect_identical(new$array.index, attr(new, "JSON") %>% purrr::flatten_int())
+  expect_identical(new$array.index, json_get(new) %>% purrr::flatten_int())
 
 })
 
@@ -530,20 +636,20 @@ test_that("bind_rows works with tbl_json", {
       name = jstring("name"),
       age = jnumber("age"))
   
-  z <- people_df %>% bind_rows(people_df)
+  z <- people_df %>% dplyr::bind_rows(people_df)
   
 
-  expect_is(attr(z,'JSON'),'list')
+  expect_is(json_get(z),'list')
   expect_is(z, 'tbl_json')
   expect_equal(nrow(z), nrow(people_df) * 2)
-  expect_equal(length(attr(z,'JSON')), nrow(people_df) * 2)
+  expect_equal(length(json_get(z)), nrow(people_df) * 2)
 })
 
 test_that("bind_rows falls back to normal behavior if not tbl_json", {
   a <- dplyr::tibble(a=c(1,2), b=c('one','two'))
   c <- dplyr::tibble(a=c(3,4), b=c('three','four'))
   
-  out <- bind_rows(a,c)
+  out <- dplyr::bind_rows(a,c)
   expect_equal(nrow(out), nrow(a) + nrow(c))
   expect_equal(names(out), c('a','b'))
   expect_is(out,'tbl_df')
@@ -581,7 +687,7 @@ test_that('dplyr::rename works with programming', {
   
   f <- json %>% spread_all %>% dplyr::rename(!!!v)
   
-  expect_identical(names(f),c('document.id','firstName','lastName'))
+  expect_identical(names(f),c('document.id','firstName','lastName','..JSON'))
   expect_is(f,'tbl_json')
 })
 
@@ -591,7 +697,7 @@ test_that('dplyr::select works with programming', {
   
   f <- json %>% spread_all %>% dplyr::select(!!!v)
   
-  expect_identical(names(f),c('Hill','valley'))
+  expect_identical(names(f),c('Hill','valley','..JSON'))
   expect_is(f,'tbl_json')
 })
 
@@ -613,7 +719,7 @@ test_that('dplyr::transmute works with programming', {
   
   f <- json %>% spread_all %>% dplyr::transmute(!!!v)
   
-  expect_identical(names(f), 'firstName')
+  expect_identical(names(f), c('firstName','..JSON'))
   expect_is(f,'tbl_json')
 })
 
@@ -629,3 +735,39 @@ test_that('dplyr::slice works with programming', {
   expect_identical(f$id,7)
   expect_is(f,'tbl_json')
 })
+
+context('json_get')
+
+test_that('json_get works', {
+  tj <- as.tbl_json('{"a": "b"}')
+  
+  expect_identical(json_get(tj), list(list("a" = "b")))
+})
+
+test_that('json_get handles error cases', {
+  tjn <- as.tbl_json('{"a": "b"}')
+  tjn[["..JSON"]] <- NULL
+  expect_identical(json_get(tjn), NULL)
+  
+  expect_identical(json_get(data.frame()), NULL)
+  expect_identical(json_get(data.frame(`..JSONA` = 1)), NULL)
+})
+
+context('json_get_column')
+
+test_that('json_get_column works', {
+  tj <- as.tbl_json('{"a": "b", "b": [1,2,3]}')
+  
+  expect_identical(
+    json_get_column(tj)[["json"]],
+    json_get(tj)
+  )
+  expect_identical(
+    json_get_column(tj, "alt")[["alt"]],
+    json_get(tj)
+  )
+  expect_identical(
+    json_get_column(tj, something.special)[["something.special"]],
+    json_get(tj)
+  )
+})
